#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "pixelscale.h"
typedef unsigned char uchar;

typedef struct line_info
{	int size;
	int left;
	int weight_l;
	int weight_r;
	int main_start;
	int main_end;
	int main_width;
} line_info;

enum NEIGHBOR
{	N_TOP, N_BOTTOM, N_LEFT, N_RIGHT  };

typedef struct neighbor
{	uchar* color;
	signed weight;
} neighbor;

static void nb_fill( uchar* color, signed weight, neighbor* nb  )
{	nb->color  = color;
	nb->weight = weight;
}

typedef struct cell_info
{	neighbor neighbors[4];
	uchar *pixel_in;
} cell_info;

typedef struct info
{	ps_size       size_in;
	ps_format_in  fmt_in;
	ps_format_out fmt_out;
	uchar         comp_n;
	double        scale_x;
	double        scale_y;
	line_info*    rows_info;
	line_info*    cols_info;
	cell_info*    cells; 
	uchar*        colors; // Contains pre-interpreted colors
	uchar        *ipl,*ipr;
	uchar*        bound_cols[4];
	char          nb;
	char          dw;
	char          dh;
} info;

typedef struct junct_info
{	uchar   low;
	uchar   high;
	double  diff_lo;
	double  diff_hi;
	double  middle;
} junct_info;

static junct_info get_junct_info( double scale, char nb )
{	junct_info ji;
	ji.low     = floor(scale);
	ji.high    = ji.low + 1;
	ji.diff_hi =  (double)ji.high - scale;
	ji.diff_lo = -(double)ji.low  + scale;
	if( nb )
	{	ji.middle = 0.0;  }
	else
	{	ji.middle = 0.0;  }
	return ji;
}

static void li_filldim( line_info *pli, int left )
{	pli->left       = left;
	pli->main_start = left;
	pli->main_end   = left + pli->size - 1;

	if( pli->weight_l < 256 )
	{	pli->main_start += 1;  }
	if( pli->weight_r < 256 )
	{	pli->main_end   -= 1;  }

	pli->main_width = pli->main_end - pli->main_start + 1;
}

static void li_setsize( junct_info ji, line_info* pli, double *diff )
{	if( *diff > ji.middle )
	{	*diff = *diff - ji.diff_hi;
		pli->size = ji.high;
	}
	else
	{	*diff = *diff + ji.diff_lo;
		pli->size = ji.low;
	}
}

static line_info *li_get( unsigned length, double scale, double softness )
{	double diff = 0.0;
	int i;
	int edgeDiff;
	char nb = (softness == 0.0);
	double dweight;
	int*   edgeWeight;
	int left = 0;
	int dummy;
	junct_info ji = get_junct_info( scale, nb );
	line_info* res = (line_info*)calloc( length, sizeof( line_info ) ); 
	line_info* pli = res;
	for( i = 0; i < length; i++ )
	{	res[i].weight_l = 256; res[i].weight_r = 256;  }
	for( i = 0; i < length; i++ )
	{	diff = scale * i - left;
		li_setsize( ji, pli, &diff );
		if( !nb )
		{	if( diff  > 0.0 )
			{	dweight = 1.0 - diff;
				if( i < length - 1 )
				{	edgeWeight = &(pli + 1)->weight_l;  }
				else
				{	edgeWeight = &dummy;  }
			}
			else
			{	dweight = 1.0 + diff;
				edgeWeight = &pli->weight_r;
			}
			dweight     = pow( dweight, softness );
			*edgeWeight = round( dweight * 256 );
		}

		li_filldim( pli, left );

		left = left + pli->size;
		pli = pli + 1;
	}
	return res;
}

static uchar* alloc_colors( info* si )
{	return (uchar*)calloc( si->size_in.h * si->size_in.w, si->comp_n );  }

static uchar* new_color( info* si )
{	return (uchar*)malloc( si->comp_n );  } 

static void init_cells( info* si )
{	int y, x;
	line_info ri, ci;
	ps_size size_in = si->size_in;
	si->cells    = (cell_info*)calloc( sizeof( cell_info ), size_in.w * size_in.h );
	uchar* col = si->colors;
	cell_info* c = si->cells;
	neighbor nb;
	for( y = 0; y < size_in.h; y++ )
	{	ri = si->rows_info[y];
		for( x = 0; x < size_in.w; x++ )
		{	ci = si->cols_info[x];
			c->pixel_in  = col;
			// TODO: Range checking is based on weight, which is rather feeble.
			nb_fill( col - si->comp_n*size_in.w, ri.weight_l, &c->neighbors[N_TOP   ] );
			nb_fill( col + si->comp_n*size_in.w, ri.weight_r, &c->neighbors[N_BOTTOM] );
			nb_fill( col - si->comp_n,           ci.weight_l, &c->neighbors[N_LEFT  ] );
			nb_fill( col + si->comp_n,           ci.weight_r, &c->neighbors[N_RIGHT ] );

			col = col + si->comp_n;
			c = c + 1;
		}
	}
}

static float roundfl( float f )
{	float f0 = floor(f);
	float f1 = f0 + 1.0;
	if( f/f0 > f1/f  )
	{	return f1; }
	else
	{	return f0;  }
}

static void get_perfect_scale_asp( float ar_in, float a_in, float a_out, float ratio, int *scalex, int *scaley )
{	const float ASPECT_IMPORTANCE = 1.14; // Heruistic parameter
	int sx, sy, sx_best, sy_best;
	int sx_max = floor(ratio);
	int sy_max = floor(ratio*a_out / ar_in );
	if( sy_max == 0 )
	{	sy_max = 1;  } // HACK for small output sizes
	float bestfit = -99.0;

	float fit;
	float err_aspect;
	for( sx = sx_max; sx > 0; sx-- )
	{	sy = roundfl( (float)sx * a_in );
		if( sy == 0 )
		{	sy = 1;  } // HACK for small output sizes
		if( sy > sy_max )
		{	sy = sy - 1;  }
		err_aspect = (float)sy/(float)sx / a_in;
		if( err_aspect < 1.0 )
		{	err_aspect = 1.0 / err_aspect;  }
		err_aspect = powf( err_aspect, ASPECT_IMPORTANCE );
		float xr = (float)sx/sx_max;
		float yr = (float)sy/sy_max;
		float sz;
		if( xr > yr )
		{	sz = xr;  }
		else
		{	sz = yr;  }
		fit = (float)(sz + 0.2) / err_aspect;
		if( fit > bestfit )
		{	bestfit = fit;
			*scalex = sx;
			*scaley = sy;
		}
	}
}

static void get_perfect_scale( ps_size size_in, float a_in, ps_size size_out, int* scalex, int* scaley )
{	float a_x, a_y;
	float ca_in, ca_out;
	float cratio; 
	float ar;
	int *sw, *sh;
	if( a_in > 1.0 )
	{	a_y = a_in;     a_x = 1.0;  }
	else
	{	a_x = 1.0/a_in; a_y = 1.0;  }
	if( size_out.w / (size_in.w * a_x) < size_out.h / (size_in.h * a_y) )
	{	ca_in  = a_in;
		ca_out = (float)size_out.h/size_out.w;
		cratio = (float)size_out.w/size_in.w;
		sw     = scalex; sh = scaley;
		ar     = (float)size_in.h/size_in.w;
	}
	else
	{	ca_in  = 1.0/a_in;
		ca_out = (float)size_out.w/size_out.h;
		cratio = (float)size_out.h/size_in.h;
		sw     = scaley; sh = scalex;
		ar     = (float)size_in.w/size_in.h;
	}
	get_perfect_scale_asp( ar, ca_in, ca_out, cratio, sw, sh );
}

static void free_if_not_null( void *ptr )
{	if( ptr != NULL )
	{	free( ptr );  }
}

static void init_bound_cols( info* si )
{	uchar i;
	for( i = 0; i < 4; i++ )
	{	si->bound_cols[i] = new_color( si );  }
}

static void free_bound_cols( info* si )
{	uchar i;
	for( i = 0; i < 4; i++ )
	{	free_if_not_null( si->bound_cols[i] );  }
}

static ps_info new_info
(	ps_format_in  fmt_in,
	ps_size       size_in,
	ps_format_out fmt_out,
	ps_size       size_out,
	uchar         comp_n,
	double        softness,
	char          dw,
	char          dh
)
{	ps_info res = (ps_info)malloc(sizeof(info));
	info* si = (info*)res;
	si->dw        = dw;
	si->dh        = dh;
	si->comp_n    = comp_n;
	si->fmt_in    = fmt_in;
	si->size_in   = size_in;
	si->fmt_out   = fmt_out;
	si->colors    = alloc_colors( si );
	si->scale_x   = ( double )size_out.w / size_in.w;
	si->scale_y   = ( double )size_out.h / size_in.h;
	si->nb        = softness == 0.0 ||
	                (	si->scale_x == (int)si->scale_x &&
						 	si->scale_y == (int)si->scale_y
	                );
	si->rows_info = li_get( size_in.h, si->scale_y, softness );
	si->cols_info = li_get( size_in.w, si->scale_x, softness );
	si->ipl       = new_color( si );
	si->ipr       = new_color( si );
	// TODO: In case of allocation errors, call ps_free and return NULL
	init_cells( si );
	init_bound_cols( si );
	return res;
}

static void handle_dwh_new( char dw, char dh, ps_format_in *fmt_in, ps_size* size, double* par )
{	if( dh )
	{	size->h /= 2;
		*par    *= 2;
	}
	if( dw )
	{	size->w      /= 2;
		*par         /= 2;
		fmt_in->step *= 2;
	}
}

static void handle_dwh_scale( char dw, char dh, ps_pixels *pix_in, ps_rect* rect )
{	if( dh )
	{	pix_in->pitch *= 2;
		rect->h       /= 2;
		rect->y       /= 2;
	}
	if( dw )
	{	rect->w /= 2;
		rect->x /= 2;
	}
}

ps_info ps_new_soft
(	ps_format_in  const fmt_in,
	ps_size       const size_in,
	ps_format_out const fmt_out,
	ps_size       const size_out,
	char          const dw,
	char          const dh,
	unsigned char const comp_n,
	double        const softness
)
{	double par_dummy;
	ps_size size_in_calc = size_in;
	ps_format_in fmt_in_calc = fmt_in;
	handle_dwh_new( dw, dh, &fmt_in_calc, &size_in_calc, &par_dummy );
	return new_info( fmt_in_calc, size_in_calc, fmt_out, size_out, comp_n, softness, dw, dh );
}

ps_info ps_new_nn
(	ps_format_in  const fmt_in,
	ps_size       const size_in,
	ps_format_out const fmt_out,
	ps_size       const size_out,
	char          const dw,
	char          const dh,
	unsigned char const comp_n
)
{	double par_dummy;
	ps_size size_in_calc = size_in;
	ps_format_in fmt_in_calc = fmt_in;
	handle_dwh_new( dw, dh, &fmt_in_calc, &size_in_calc, &par_dummy );
	return new_info( fmt_in_calc, size_in_calc, fmt_out, size_out, comp_n, 0.0, dw, dh );  }

ps_info ps_new_perfect
(	ps_format_in  const fmt_in,
	ps_size       const size_in,
	ps_format_out const fmt_out,
	ps_size       const size_out,
	char          const dw,
	char          const dh,
	unsigned char const comp_n,
	double        const aspect_in,
	ps_size*      const size_res
)
{	int sx, sy;
	double aspect = aspect_in;
	ps_size size_in_calc = size_in;
	ps_format_in fmt_in_calc = fmt_in;
	handle_dwh_new( dw, dh, &fmt_in_calc, &size_in_calc, &aspect );
	get_perfect_scale( size_in_calc, aspect, size_out, &sx, &sy );
	size_res->w = sx * size_in_calc.w;
	size_res->h = sy * size_in_calc.h;
	return new_info( fmt_in_calc, size_in_calc, fmt_out, *size_res, comp_n, 0.0, dw, dh ); 
}

void ps_free ( ps_info const osi )
{	info* si = (info*)osi;
	free_if_not_null( si->ipl );
	free_if_not_null( si->ipr );
	free_if_not_null( si->cols_info );
	free_if_not_null( si->rows_info );
	free_if_not_null( si->colors );
	free_if_not_null( si->cells );
	free_bound_cols( si );
	free( si );
}

// TODO: Maybe precalculate the difference 256 - weight also?
static inline uchar interpolate_comp( unsigned weight, uchar current, uchar that )
{	return ( (int)current * weight + (int)that * ( 256 - weight ) ) >> 8;  }

static inline void interpolate_pixel( unsigned weight, uchar* current, uchar* that, int count, uchar* res )
{	int c;
	for( c = 0; c < count; c++ )
	{	*res = interpolate_comp( weight, *current, *that );
		res++; current++; that++;
	}
}

static inline void get_boundary_colors( cell_info *cell, line_info *rowInfo, line_info *colInfo, info* si  )
{	uchar i;
	neighbor* nb;
	for( i = 0; i < 4; i++ )
	{	nb = &cell->neighbors[i];
		if( nb->weight < 256 )
		{	interpolate_pixel( nb->weight, cell->pixel_in, nb->color, si->comp_n, si->bound_cols[i] );  }
	}
}

static inline uchar* addr( uchar* start, int size, int x, int y, int pitch )
{	return start + y * pitch + x * size;  }

// TODO: restructure the internal types so as to make this easier:
static inline uchar* addr_img( uchar step, ps_pixels pixels, int x, int y )
{	return addr( pixels.pixels, step, x, y, pixels.pitch );  }

static inline void get_pixel( ps_format_in fmt, uchar* whence, uchar* pixel )
{	if( fmt.get_pixel == NULL )
	{	memcpy( pixel, whence, fmt.bpp );  }
	else
	{	fmt.get_pixel( whence, pixel );  }
}

static void get_colors( info* si, ps_pixels pixels, ps_rect rect )
{	int x,y;
	uchar *px_rs, *px_in; 
	uchar *col, *col_rs;
	col_rs = si->colors + si->comp_n * (rect.y * si->size_in.w + rect.x); // Use getaddress for this as well.
	px_rs = addr_img( si->fmt_in.step, pixels, rect.x, rect.y );
	for( y = rect.y; y < rect.y + rect.h; y++ )
	{	px_in = px_rs;
		col   = col_rs;
		for( x = rect.x; x < rect.x + rect.w; x++ )
		{	get_pixel( si->fmt_in, px_in, col ); 
			col = col + si->comp_n;
			px_in = px_in + si->fmt_in.step;
		}
		col_rs += si->size_in.w * si->comp_n;
		px_rs  += pixels.pitch;
	}
}

static inline void put_pixel( ps_format_out fmt, uchar* whither, uchar* pixel )
{	if( fmt.put_pixel == NULL )
	{	memcpy( whither, pixel, fmt.bpp );  }
	else
	{	fmt.put_pixel( whither, pixel );  }
}

static inline void put_and_shift( ps_format_out fmt, uchar** row, uchar* pixel )
{	put_pixel( fmt, *row, pixel );
	*row = *row + fmt.step; 
}

static inline void fill_subrow( uchar** row, line_info *ci, uchar* left, uchar* right, uchar* current, ps_format_out fmt_out )
{	uchar i;
	if( ci->weight_l < 256 )
	{	put_and_shift( fmt_out, row, left );  }
	for( i = 0; i < ci->main_width; i++ )
	{	put_and_shift( fmt_out, row, current );  }
	if( ci->weight_r < 256 )
	{	put_and_shift( fmt_out, row, right );  }
}

static inline uchar color_diff( uchar* a, uchar* b, uchar count )
{	unsigned full = 0;
	int comp_diff;
	uchar i;
	for( i = 0; i < count; i++ )
	{	comp_diff = *a - *b;
		if( comp_diff < 0 )
		{	comp_diff = - comp_diff;  }
		full = full + comp_diff;
		a = a + 1;
		b = b + 1;
	}
	return (uchar)(full / count);
}

static inline void interpolate_corner( uchar comp_n, uchar* current, uchar cur_diff, uchar add_diff, uchar* other, uchar* res )
{	double dweight = (double)cur_diff / ( (double)add_diff + (double)cur_diff ) * 255;
	uchar   weight = ( uchar )round( dweight );
	interpolate_pixel( weight, current, other, comp_n, res );
}

static inline void fill_boundary(	uchar** row, info* si, line_info* ci, NEIGHBOR n, cell_info *cell )
{	uchar cur_diff, add_diff;
	uchar weight;
	uchar    **bounds    = si->bound_cols;
	neighbor  *neighbors = cell->neighbors;
	cur_diff = color_diff( neighbors[n].color, cell->pixel_in, si->comp_n );
	if( ci->weight_l < 256 )
	{	add_diff = color_diff( neighbors[N_LEFT].color, cell->pixel_in, si->comp_n );
		interpolate_corner( si->comp_n, bounds[n], cur_diff, add_diff, bounds[N_LEFT], si->ipl ); 
	}
	if( ci->weight_r < 256 )
	{	add_diff = color_diff( neighbors[N_RIGHT].color, cell->pixel_in, si->comp_n );
		interpolate_corner( si->comp_n, bounds[n], cur_diff, add_diff, bounds[N_RIGHT], si->ipr ); 
	}
	fill_subrow( row, ci, si->ipl, si->ipr, bounds[n], si->fmt_out );
}

static void inc_rect( ps_rect* rect, ps_size size, char l, char r, char t, char b )
{	if( ( rect->x + rect->w < size.w ) && r )
	{	rect->w++;  }
	if( rect->y + rect->h < size.h && b )
	{	rect->h++;  }
	if( rect->x > 0 && l )
	{	rect->x--; rect->w++;  }
	if( rect->y > 0 && t )
	{	rect->y--; rect->h++;  }
}

static ps_rect get_color_rect( info* si, ps_rect rect, ps_size size )
{	ps_rect colRect = rect;
	inc_rect
	(	&colRect, size, 
		si->cols_info[rect.x].weight_l < 256,
		si->cols_info[rect.x + rect.w - 1].weight_r < 256,
		si->rows_info[rect.y].weight_l < 256,
		si->rows_info[rect.y + rect.h - 1].weight_r < 256
	);
	return colRect;
}

void ps_scale( ps_info const osi, ps_pixels pix_in, ps_pixels const pix_out, ps_rect* area )
{	info* si = (info*)osi;
	int x_in, y_in, x_pix, y_pix;
	uchar mid_y;
	line_info ri;
	line_info ci;
	uchar* mid_row;

	handle_dwh_scale( si->dw, si->dh, &pix_in, area );
	if( !si->nb )
	{	inc_rect( area, si->size_in, 1, 1, 1, 1 );  }

	mid_row    = (uchar*)malloc( pix_out.pitch );
	ps_rect colRect = get_color_rect( si, *area, si->size_in );

	// TODO: Determine it in the constructor and dispatch via func. pointers.
	char through = ( si->nb && si->fmt_out.put_pixel == NULL && si->fmt_in.get_pixel == NULL );
	if( !through )
	{	get_colors( si, pix_in, colRect );  }

	cell_info *cell, *cell_rs = si->cells + area->y * si->size_in.w + area->x;
	uchar* midRow_out;
	uchar  *top_cur, *bot_cur, *mid_cur;
	uchar* outRow0 = addr_img
	(	si->fmt_out.step, pix_out, si->cols_info[area->x].left, si->rows_info[area->y].left );
	unsigned width_bytes = si->fmt_out.step * 
		( si->cols_info[area->x + area->w - 1].left + si->cols_info[area->x+area->w-1].size -
		  si->cols_info[area->x].left
		);
	for( y_in = area->y; y_in < area->y + area->h; y_in++ )
	{	ri = si->rows_info[y_in];
		if( !si->nb ) // Little effect on speed
		{	top_cur = outRow0;
			bot_cur = outRow0 + (ri.size - 1) * pix_out.pitch;
		}
		mid_cur = mid_row;
		cell    = cell_rs;
		if( ri.main_width > 0 ) 
		{	midRow_out = outRow0 + ( ri.main_start - ri.left ) * pix_out.pitch;  }
		for( x_in = area->x; x_in < area->x + area->w; x_in++ )
		{	ci = si->cols_info[x_in];
			if( si->nb )
			{	goto fill_main;  }
			get_boundary_colors( cell, &ri, &ci, si );
			if( ri.weight_l  < 256 )
			{	fill_boundary( &top_cur, si, &ci, N_TOP,    cell );  }
			if( ri.weight_r  < 256 )
			{	fill_boundary( &bot_cur, si, &ci, N_BOTTOM, cell );  }
			if( ri.main_width == 0 )
			{	goto next_pix;  }
		fill_main:
			if( !through )
			{	fill_subrow
				(	&mid_cur, &ci,
					si->bound_cols[N_LEFT ],
					si->bound_cols[N_RIGHT],
					cell->pixel_in, si->fmt_out
				); 
			}
			else
			{	uchar* src = addr_img( si->fmt_in.step, pix_in, x_in, y_in );
				for( int mc = 0; mc < ci.size; mc++ )
				{	memcpy( mid_cur, src, si->fmt_in.bpp );
					mid_cur = mid_cur + si->fmt_out.step;
				}
			}
		next_pix:
			cell += 1;
		}
		for( mid_y = 0; mid_y < ri.main_width; mid_y++ )
		{	memcpy( midRow_out, mid_row, width_bytes );
			midRow_out = midRow_out + pix_out.pitch;
		}
		outRow0 = outRow0 + ri.size * pix_out.pitch;
		cell_rs = cell_rs + si->size_in.w;
	}
	free(mid_row);
	area->x = round( si->scale_x * area->x );
	area->w = round( si->scale_x * area->w );
	area->y = round( si->scale_y * area->y );
	area->h = round( si->scale_y * area->h );
}
